@@ -7,7 +7,7 @@ __license__  = "GNU GPL version 3 or any later version"
 
 # Modified by Kristian B. Oelgaard 2010
 # Modified by Marie E. Rognes 2010
-# Last changed: 2010-02-12
+# Last changed: 2010-03-11
 
 # Python modules
 import re, numpy, platform
@@ -34,7 +34,7 @@ format.update({
     "comment":        lambda v: "// %s" % v,
     "if":             lambda c, v: "if (%s)\n{\n%s\n}\n" % (c, v),
     "loop":           lambda i, j, k: "for (unsigned int %s = %s; %s < %s; %s++)"% (i, j, i, k, i),
-    "generate loop":  lambda v, w: _generate_loop(v, w),
+    "generate loop":  lambda v, w, _indent=0: _generate_loop(v, w, _indent),
     "is equal":       " == ",
     "do nothing":     "// Do nothing"
 })
@@ -69,6 +69,10 @@ format.update({
     "ln":             lambda v: "std::log(%s)" % str(v),
     "cos":            lambda v: "std::cos(%s)" % str(v),
     "sin":            lambda v: "std::sin(%s)" % str(v),
+    "tan":            lambda v: "std::tan(%s)" % str(v),
+    "acos":           lambda v: "std::acos(%s)" % str(v),
+    "asin":           lambda v: "std::asin(%s)" % str(v),
+    "atan":           lambda v: "std::atan(%s)" % str(v),
     "absolute value": lambda v: "std::abs(%s)" % str(v),
     "sqrt":           lambda v: "std::sqrt(%s)" % str(v),
     "addition":       lambda v: _add(v),
@@ -98,6 +102,8 @@ format.update({
     "x coordinate":     "X",
     "y coordinate":     "Y",
     "z coordinate":     "Z",
+    "ip coordinates":   lambda i, j: "X%d[%d]" % (i, j),
+    "affine map table": lambda i, j: "FEA%d_f%d" % (i, j),
     "coordinates":      "x"
 })
 
@@ -162,7 +168,12 @@ format.update({
     "integration points": "ip",
     "first free index":   "j",
     "second free index":  "k",
-    "geometry constant":  "G",
+    "geometry constant":  lambda i: "G[%d]" % i,
+    "ip constant":        lambda i: "I[%d]" % i,
+    "basis constant":     lambda i: "B[%d]" % i,
+#    "geometry constant":  lambda i: "G%d" % i,
+#    "ip constant":        lambda i: "I%d" % i,
+#    "basis constant":     lambda i: "B%d" % i,
     "function value":     lambda i: "F%d" % i,
     "nonzero columns":    lambda i: "nzc%d" % i,
     "weight":             lambda i: "W%d" % (i),
@@ -194,6 +205,8 @@ format.update({
     "facet determinant":    lambda n, r=None: facet_determinant[n] % {"restriction": choose_map[r]},
     "fiat coordinate map":  lambda n: fiat_coordinate_map[n],
     "generate normal":      lambda d, i: _generate_normal(d, i),
+    "generate ip coordinates":  lambda g, num_ip, name, ip, r=None: (ip_coordinates[g][0], ip_coordinates[g][1] % \
+                                {"restriction": choose_map[r], "ip": ip, "name": name, "num_ip": num_ip}),
     "scale factor snippet": scale_factor,
     "map onto physical":    map_onto_physical,
     "combinations":         combinations_snippet,
@@ -240,7 +253,7 @@ def _delete_array(name, size=None):
     if size is None:
         return "delete [] %s;" % name
     f_r = format["free indices"][0]
-    code = _generate_loop(["delete [] %s;" % format["component"](name, f_r)], [(f_r, 0, size)])
+    code = format["generate loop"](["delete [] %s;" % format["component"](name, f_r)], [(f_r, 0, size)])
     code.append("delete [] %s;" % name)
     return "\n".join(code)
 
@@ -404,7 +417,7 @@ def _tabulate_tensor(vals):
     else:
         error("Not an N-dimensional array:\n%s" % tensor)
 
-def _generate_loop(lines, loop_vars):
+def _generate_loop(lines, loop_vars, _indent):
     "This function generates a loop over a vector or matrix."
 
     # Prefetch formats to speed up code generation.
@@ -417,7 +430,6 @@ def _generate_loop(lines, loop_vars):
         return lines
 
     code = []
-    _indent = 0
     for ls in loop_vars:
         # Get index and lower and upper bounds.
         index, lower, upper = ls
