@@ -6,7 +6,7 @@ __copyright__ = "Copyright (C) 2009-2010 Kristian B. Oelgaard"
 __license__  = "GNU GPL version 3 or any later version"
 
 # Modified by Anders Logg, 2009
-# Last changed: 2010-02-08
+# Last changed: 2010-03-11
 
 # Python modules.
 from numpy import shape
@@ -19,6 +19,7 @@ from ufl.classes import FixedIndex
 from ufl.classes import IntValue
 from ufl.classes import FloatValue
 from ufl.classes import Coefficient
+from ufl.expr import Operator
 
 # UFL Algorithms.
 from ufl.algorithms.printing import tree_format
@@ -37,10 +38,10 @@ from ffc.quadrature.symbolics import create_float, create_symbol, create_product
 class QuadratureTransformerOpt(QuadratureTransformerBase):
     "Transform UFL representation to quadrature code."
 
-    def __init__(self, ir, optimise_parameters):
+    def __init__(self, *args):
 
         # Initialise base class.
-        QuadratureTransformerBase.__init__(self, ir, optimise_parameters)
+        QuadratureTransformerBase.__init__(self, *args)
 #        set_format(format)
 
     # -------------------------------------------------------------------------
@@ -153,15 +154,19 @@ class QuadratureTransformerOpt(QuadratureTransformerBase):
             return {(): create_product([val]*expo.value())}
         elif isinstance(expo, FloatValue):
             exp = format["floating point"](expo.value())
-            sym = create_symbol(format["std power"](str(val), exp), val.t)
-            sym.base_expr = val
-            sym.base_op = 1 # Add one operation for the pow() function.
+#            sym = create_symbol(format["std power"](str(val), exp), val.t)
+#            sym.base_expr = val
+#            sym.base_op = 1 # Add one operation for the pow() function.
+            sym = create_symbol(format["std power"], val.t, val, 1)
+            sym.exp = exp
             return {(): sym}
-        elif isinstance(expo, Coefficient):
+        elif isinstance(expo, (Coefficient, Operator)):
             exp = self.visit(expo)
-            sym = create_symbol(format["std power"](str(val), exp[()]), val.t)
-            sym.base_expr = val
-            sym.base_op = 1 # Add one operation for the pow() function.
+#            sym = create_symbol(format["std power"](str(val), exp[()]), val.t)
+#            sym.base_expr = val
+#            sym.base_op = 1 # Add one operation for the pow() function.
+            sym = create_symbol(format["std power"], val.t, val, 1)
+            sym.exp = exp[()]
             return {(): sym}
         else:
             error("power does not support this exponent: " + repr(expo))
@@ -175,9 +180,10 @@ class QuadratureTransformerOpt(QuadratureTransformerBase):
 
         # Take absolute value of operand.
         val = operands[0][()]
-        new_val = create_symbol(format["absolute value"](str(val)), val.t)
-        new_val.base_expr = val
-        new_val.base_op = 1 # Add one operation for taking the absolute value.
+#        new_val = create_symbol(format["absolute value"](str(val)), val.t)
+#        new_val.base_expr = val
+#        new_val.base_op = 1 # Add one operation for taking the absolute value.
+        new_val = create_symbol(format["absolute value"], val.t, val, 1)
         return {():new_val}
 
     # -------------------------------------------------------------------------
@@ -189,11 +195,25 @@ class QuadratureTransformerOpt(QuadratureTransformerBase):
         # Get the component
         components = self.component()
 
-        # Safety checks.
+        # Safety check.
         ffc_assert(not operands, "Didn't expect any operands for FacetNormal: " + repr(operands))
-        ffc_assert(len(components) == 1, "FacetNormal expects 1 component index: " + repr(components))
 
-        normal_component = format["normal component"](self.restriction, components[0])
+        # Handle 1D as a special case.
+        # FIXME: KBO: This has to change for mD elements in R^n : m < n
+        if self.geo_dim == 1:
+            # Safety check.
+            ffc_assert(len(components) == 0, "FacetNormal in 1D does not expect a component index: " + repr(components))
+            normal_component = format["normal component"](self.restriction, "")
+            self.trans_set.add(normal_component)
+        else:
+
+            # Safety check.
+            ffc_assert(len(components) == 1, "FacetNormal expects 1 component index: " + repr(components))
+
+            # We get one component.
+            normal_component = format["normal component"](self.restriction, components[0])
+            self.trans_set.add(normal_component)
+
         return {(): create_symbol(normal_component, GEO)}
 
     def create_argument(self, ufl_argument, derivatives, component, local_comp,
@@ -356,9 +376,10 @@ class QuadratureTransformerOpt(QuadratureTransformerBase):
         # Use format function on value of operand.
         operand = operands[0]
         for key, val in operand.items():
-            new_val = create_symbol(format_function(str(val)), val.t)
-            new_val.base_expr = val
-            new_val.base_op = 1 # Add one operation for the math function.
+#            new_val = create_symbol(format_function(str(val)), val.t)
+#            new_val.base_expr = val
+#            new_val.base_op = 1 # Add one operation for the math function.
+            new_val = create_symbol(format_function, val.t, val, 1)
             operand[key] = new_val
         return operand
 
@@ -390,16 +411,5 @@ class QuadratureTransformerOpt(QuadratureTransformerBase):
         ops = self._count_operations(value)
         used_psi_tables = set([self.psi_tables_map[b] for b in value.get_unique_vars(BASIS)])
 
-        return [value, ops, trans_set, used_points, used_psi_tables]
-
-#        value = optimise_code(value, self.ip_consts, self.geo_consts, self.trans_set)
-
-        # Check if value is zero
-#        if not value.val:
-#            zero = True
-#        # Update the set of used psi tables through the name map if the value is not zero.
-#        else:
-#            self.used_psi_tables.update([self.psi_tables_map[b] for b in value.get_unique_vars(BASIS)])
-
-#        return value
+        return (value, ops, [trans_set, used_points, used_psi_tables])
 
