@@ -5,7 +5,8 @@ __date__ = "2009-01-07"
 __copyright__ = "Copyright (C) 2009-2010 Kristian B. Oelgaard"
 __license__  = "GNU GPL version 3 or any later version"
 
-# Last changed: 2010-02-11
+# Modified by Mehdi Nikbakht, 2010
+# Last changed: 2010-06-15
 
 # Python modules.
 import numpy
@@ -28,6 +29,8 @@ def generate_integral_code(ir, prefix, parameters):
     code["classname"] = format["classname " + ir["domain_type"] + "_integral"](prefix, ir["form_id"], ir["domain_id"])
     code["members"] = ""
     code["constructor"] = format["do nothing"]
+    code["constructor_arguments"] = ""
+    code["initializer_list"] = ""
     code["destructor"] = format["do nothing"]
     code["tabulate_tensor"] = _tabulate_tensor(ir, parameters)
 
@@ -38,7 +41,7 @@ def _tabulate_tensor(ir, parameters):
 
     f_comment       = format["comment"]
     f_G             = format["geometry constant"]
-    f_const_double  = format["const float declaration"]
+    f_const_double  = format["assign"]
     f_switch        = format["switch"]
     f_float         = format["float"]
     f_assign        = format["assign"]
@@ -49,13 +52,13 @@ def _tabulate_tensor(ir, parameters):
     f_facet         = format["facet"]
 
     # Get data.
-    optimise_parameters = ir["optimise_parameters"]
-    domain_type         = ir["domain_type"]
-    geometric_dimension = ir["geometric_dimension"]
-    num_facets          = ir["num_facets"]
-    prim_idims          = ir["prim_idims"]
-    integrals           = ir["trans_integrals"]
-    geo_consts          = ir["geo_consts"]
+    opt_par     = ir["optimise_parameters"]
+    domain_type = ir["domain_type"]
+    geo_dim     = ir["geometric_dimension"]
+    num_facets  = ir["num_facets"]
+    prim_idims  = ir["prim_idims"]
+    integrals   = ir["trans_integrals"]
+    geo_consts  = ir["geo_consts"]
 
     # Create sets of used variables.
     used_weights    = set()
@@ -64,40 +67,44 @@ def _tabulate_tensor(ir, parameters):
     trans_set       = set()
     sets = [used_weights, used_psi_tables, used_nzcs, trans_set]
 
+    affine_tables = {}
+    quadrature_weights = ir["quadrature_weights"]
+
     operations = []
     if domain_type == "cell":
         # Update treansformer with facets and generate code + set of used geometry terms.
-        tensor_code, mem_code, num_ops = _generate_element_tensor(integrals, sets, optimise_parameters)
+        tensor_code, mem_code, num_ops = _generate_element_tensor(integrals, sets, \
+                                         opt_par)
         tensor_code = "\n".join(tensor_code)
 
         # Set operations equal to num_ops (for printing info on operations).
-        operations.append(num_ops)
+        operations.append([num_ops])
 
         # Get Jacobian snippet.
         # FIXME: This will most likely have to change if we support e.g., 2D elements in 3D space.
-        jacobi_code = format["jacobian and inverse"](geometric_dimension)
+        jacobi_code = format["jacobian and inverse"](geo_dim)
         jacobi_code += "\n\n" + format["scale factor snippet"]
 
     elif domain_type == "exterior_facet":
         cases = [None for i in range(num_facets)]
         for i in range(num_facets):
             # Update treansformer with facets and generate case code + set of used geometry terms.
-            c, mem_code, ops = _generate_element_tensor(integrals[i], sets, optimise_parameters)
+            c, mem_code, ops = _generate_element_tensor(integrals[i], sets, opt_par)
             case = [f_comment("Total number of operations to compute element tensor (from this point): %d" % ops)]
             case += c
             cases[i] = "\n".join(case)
 
             # Save number of operations (for printing info on operations).
-            operations.append((i, ops))
+            operations.append([i, ops])
 
         # Generate tensor code for all cases using a switch.
         tensor_code = f_switch(f_facet(None), cases)
 
         # Get Jacobian snippet.
         # FIXME: This will most likely have to change if we support e.g., 2D elements in 3D space.
-        jacobi_code = format["jacobian and inverse"](geometric_dimension)
-        jacobi_code += "\n\n" + format["facet determinant"](geometric_dimension)
-        jacobi_code += "\n\n" + format["generate normal"](geometric_dimension, domain_type)
+        jacobi_code = format["jacobian and inverse"](geo_dim)
+        jacobi_code += "\n\n" + format["facet determinant"](geo_dim)
+        jacobi_code += "\n\n" + format["generate normal"](geo_dim, domain_type)
 
     elif domain_type == "interior_facet":
         # Modify the dimensions of the primary indices because we have a macro element
@@ -107,36 +114,37 @@ def _tabulate_tensor(ir, parameters):
         for i in range(num_facets):
             for j in range(num_facets):
                 # Update treansformer with facets and generate case code + set of used geometry terms.
-                c, mem_code, ops = _generate_element_tensor(integrals[i][j], sets, optimise_parameters)
+                c, mem_code, ops = _generate_element_tensor(integrals[i][j], sets, \
+                                                            opt_par)
                 case = [f_comment("Total number of operations to compute element tensor (from this point): %d" % ops)]
                 case += c
                 cases[i][j] = "\n".join(case)
 
                 # Save number of operations (for printing info on operations).
-                operations.append((i, j, ops))
+                operations.append([i, j, ops])
 
         # Generate tensor code for all cases using a switch.
         tensor_code = f_switch(f_facet("+"), [f_switch(f_facet("-"), cases[i]) for i in range(len(cases))])
 
         # Get Jacobian snippet.
         # FIXME: This will most likely have to change if we support e.g., 2D elements in 3D space.
-        jacobi_code  = format["jacobian and inverse"](geometric_dimension, "+")
+        jacobi_code  = format["jacobian and inverse"](geo_dim, "+")
         jacobi_code += "\n\n"
-        jacobi_code += format["jacobian and inverse"](geometric_dimension, "-")
+        jacobi_code += format["jacobian and inverse"](geo_dim, "-")
         jacobi_code += "\n\n"
-        jacobi_code += format["facet determinant"](geometric_dimension, "+")
-        jacobi_code += "\n\n" + format["generate normal"](geometric_dimension, domain_type)
+        jacobi_code += format["facet determinant"](geo_dim, "+")
+        jacobi_code += "\n\n" + format["generate normal"](geo_dim, domain_type)
     else:
         error("Unhandled integral type: " + str(integral_type))
 
     # After we have generated the element code for all facets we can remove
     # the unused transformations and tabulate the used psi tables and weights.
     common = [remove_unused(jacobi_code, trans_set)]
-    quadrature_weights = ir["quadrature_weights"]
     common += _tabulate_weights([quadrature_weights[p] for p in used_weights])
     name_map = ir["name_map"]
     tables = ir["unique_tables"]
-    common += _tabulate_psis(tables, used_psi_tables, name_map, used_nzcs, optimise_parameters)
+    tables.update(affine_tables)
+    common += _tabulate_psis(tables, used_psi_tables, name_map, used_nzcs, opt_par)
 
     # Reset the element tensor (array 'A' given as argument to tabulate_tensor() by assembler)
     # Handle functionals.
@@ -151,18 +159,23 @@ def _tabulate_tensor(ir, parameters):
     # Create the constant geometry declarations (only generated if simplify expressions are enabled).
     geo_ops, geo_code = generate_aux_constants(geo_consts, f_G, f_const_double)
     if geo_code:
+        common += [f_comment("Number of operations to compute geometry constants: %d." % geo_ops)]
+        common += [format["declaration"](format["float declaration"], f_G(len(geo_consts)))]
         common += geo_code
 
     # Add comments.
     common += ["", f_comment("Compute element tensor using UFL quadrature representation")]
-    common += [f_comment("Optimisations: %s" % ", ".join([str(i) for i in optimise_parameters.items()]))]
+    common += [f_comment("Optimisations: %s" % ", ".join([str((k, opt_par[k]))\
+                for k in sorted(opt_par.keys())]))]
 
     # Print info on operation count.
     message = {"cell":           "Cell, number of operations to compute tensor: %d",
                "exterior_facet": "Exterior facet %d, number of operations to compute tensor: %d",
                "interior_facet": "Interior facets (%d, %d), number of operations to compute tensor: %d"}
     for ops in operations:
-        info(message[domain_type] % ops)
+        # Add geo ops count to integral ops count for writing info.
+        ops[-1] += geo_ops
+        info(message[domain_type] % tuple(ops))
     return "\n".join(common) + "\n" + tensor_code
 
 def _generate_element_tensor(integrals, sets, optimise_parameters):
@@ -171,8 +184,13 @@ def _generate_element_tensor(integrals, sets, optimise_parameters):
     # Prefetch formats to speed up code generation.
     f_comment = format["comment"]
     f_ip      = format["integration points"]
-    f_Gip     = format["geometry constant"] + format["integration points"]
+    f_I       = format["ip constant"]
     f_loop    = format["generate loop"]
+    f_coords  = format["generate ip coordinates"]
+    f_double  = format["float declaration"]
+    f_decl    = format["declaration"]
+    f_X       = format["ip coordinates"]
+
 
     # Initialise return values.
     element_code     = []
@@ -186,12 +204,25 @@ def _generate_element_tensor(integrals, sets, optimise_parameters):
     members_code = ""
     # We receive a dictionary {num_points: form,}.
     # Loop points and forms.
-    for points, terms, functions, ip_consts in integrals:
+    for points, terms, functions, ip_consts, coordinate in integrals:
 
         element_code += ["", f_comment("Loop quadrature points for integral.")]
 
         ip_code = []
         num_ops = 0
+
+        # Generate code to compute coordinates if used.
+        if coordinate:
+            name, geo_dim, ip, r = coordinate
+            element_code += ["", f_comment("Declare array to hold physical coordinate of quadrature point.")]
+            element_code += [f_decl(f_double, f_X(points, geo_dim))]
+            ops, coord_code = f_coords(geo_dim, points, name, ip, r)
+            ip_code += ["", f_comment("Compute physical coordinate of quadrature point, operations: %d." % ops)]
+            ip_code += [coord_code]
+            num_ops += ops
+            # Update used psi tables.
+            sets[1].add(name)
+
         # Generate code to compute function values.
         if functions:
             func_code, ops = _generate_functions(functions, sets)
@@ -199,11 +230,14 @@ def _generate_element_tensor(integrals, sets, optimise_parameters):
             num_ops += ops
 
         # Generate code for ip constant declarations.
-        ip_const_ops, ip_const_code = generate_aux_constants(ip_consts, f_Gip,\
-                                        format["const float declaration"], True)
+#        ip_const_ops, ip_const_code = generate_aux_constants(ip_consts, f_I,\
+#                                        format["const float declaration"], True)
+        ip_const_ops, ip_const_code = generate_aux_constants(ip_consts, f_I,\
+                                        format["assign"], True)
         num_ops += ip_const_ops
         if ip_const_code:
             ip_code += ["", f_comment("Number of operations to compute ip constants: %d" %ip_const_ops)]
+            ip_code += [format["declaration"](format["float declaration"], f_I(len(ip_consts)))]
             ip_code += ip_const_code
 
         # Generate code to evaluate the element tensor.
@@ -234,7 +268,6 @@ def _generate_functions(functions, sets):
     f_decl         = format["declaration"]
     f_r            = format["free indices"][0]
     f_iadd         = format["iadd"]
-#    f_nzc          = format["nonzero columns"](0).split("0")[0]
     f_loop         = format["generate loop"]
 
     # Create the function declarations.
@@ -299,6 +332,7 @@ def _generate_integral_code(points, terms, sets, optimise_parameters):
     f_add           = format["add"]
     f_A             = format["element tensor"]
     f_loop          = format["generate loop"]
+    f_B             = format["basis constant"]
 
     # Initialise return values.
     code = []
@@ -308,16 +342,16 @@ def _generate_integral_code(points, terms, sets, optimise_parameters):
     # Extract sets.
     used_weights, used_psi_tables, used_nzcs, trans_set = sets
 
-
-    # TODO: Verify that test and trial functions will ALWAYS be rearranged to 0 and 1.
-    indices = {-2: format["first free index"], -1: format["second free index"],
-                0: format["first free index"],  1: format["second free index"]}
-
     # Loop terms and create code.
-    for key, data in terms.items():
+    for loop, (data, entry_vals) in terms.items():
+
+        # If we don't have any entry values, there's no need to generate the
+        # loop.
+        if not entry_vals:
+            continue
 
         # Get data.
-        val, ops, t_set, u_weights, u_psi_tables, u_nzcs = data
+        t_set, u_weights, u_psi_tables, u_nzcs, basis_consts = data
 
         # If we have a value, then we also need to update the sets of used variables.
         trans_set.update(t_set)
@@ -325,80 +359,35 @@ def _generate_integral_code(points, terms, sets, optimise_parameters):
         used_psi_tables.update(u_psi_tables)
         used_nzcs.update(u_nzcs)
 
-        # Compute number of operations to compute entry
-        # (add 1 because of += in assignment).
-        entry_ops = ops + 1
+        # Generate code for basis constant declarations.
+#        basis_const_ops, basis_const_code = generate_aux_constants(basis_consts, f_B,\
+#                                        format["const float declaration"], True)
+        basis_const_ops, basis_const_code = generate_aux_constants(basis_consts, f_B,\
+                                        format["assign"], True)
+        decl_code = []
+        if basis_consts:
+            decl_code = [format["declaration"](format["float declaration"], f_B(len(basis_consts)))]
+        loops[loop] = [basis_const_ops, decl_code + basis_const_code]
 
-        # Create comment for number of operations
-        entry_ops_comment = f_comment("Number of operations to compute entry: %d" % entry_ops)
+        for entry, value, ops in entry_vals:
+            # Compute number of operations to compute entry
+            # (add 1 because of += in assignment).
+            entry_ops = ops + 1
 
-        # Create appropriate entries.
-        # FIXME: We only support rank 0, 1 and 2.
-        entry = ""
-        loop = ()
-        if len(key) == 0:
-            entry = "0"
-        elif len(key) == 1:
-            key = key[0]
-            # Checking if the basis was a test function.
-            # TODO: Make sure test function indices are always rearranged to 0.
-            ffc_assert(key[0] == -2 or key[0] == 0, \
-                        "Linear forms must be defined using test functions only: " + repr(key))
-            index_j, entry, range_j, space_dim_j = key
-            loop = ((indices[index_j], 0, range_j),)
-            if range_j == 1 and optimise_parameters["ignore ones"]:
-                loop = ()
-            # Multiply number of operations to compute entries by range of loop.
-            entry_ops *= range_j
+            # Create comment for number of operations
+            entry_ops_comment = f_comment("Number of operations to compute entry: %d" % entry_ops)
 
-        elif len(key) == 2:
-            # Extract test and trial loops in correct order and check if for is legal.
-            key0, key1 = (0, 0)
-            for k in key:
-                ffc_assert(k[0] in indices, \
-                            "Bilinear forms must be defined using test and trial functions (index -2, -1, 0, 1): " + repr(k))
-                if k[0] == -2 or k[0] == 0:
-                    key0 = k
-                else:
-                    key1 = k
-            index_j, entry_j, range_j, space_dim_j = key0
-            index_k, entry_k, range_k, space_dim_k = key1
-
-            loop = []
-            if not (range_j == 1 and optimise_parameters["ignore ones"]):
-                loop.append((indices[index_j], 0, range_j))
-            if not (range_k == 1 and optimise_parameters["ignore ones"]):
-                loop.append((indices[index_k], 0, range_k))
-            entry = f_add([f_mul([entry_j, str(space_dim_k)]), entry_k])
-            loop = tuple(loop)
-
-            # Multiply number of operations to compute entries by range of loops.
-            entry_ops *= range_j*range_k
-        else:
-            error("Only rank 0, 1 and 2 tensors are currently supported: " + repr(key))
-
-        # Generate the code line for the entry.
-        # Try to evaluate entry ("3*6 + 2" --> "20").
-        try:
-            entry = str(eval(entry))
-        except:
-            pass
-
-        entry_code = f_iadd(f_A(entry), val)
-
-        if loop not in loops:
-            loops[loop] = [entry_ops, [entry_ops_comment, entry_code]]
-        else:
+            entry_code = f_iadd(f_A(entry), value)
             loops[loop][0] += entry_ops
             loops[loop][1] += [entry_ops_comment, entry_code]
 
     # Write all the loops of basis functions.
     for loop, ops_lines in loops.items():
         ops, lines = ops_lines
-
+        prim_ops = reduce(lambda i, j: i*j, [ops] + [l[2] for l in loop])
         # Add number of operations for current loop to total count.
-        num_ops += ops
-        code += ["", f_comment("Number of operations for primary indices: %d" % ops)]
+        num_ops += prim_ops
+        code += ["", f_comment("Number of operations for primary indices: %d" % prim_ops)]
         code += f_loop(lines, loop)
 
     return code, num_ops
@@ -535,7 +524,7 @@ def _tabulate_psis(tables, used_psi_tables, inv_name_map, used_nzcs, optimise_pa
             code += [f_decl(f_table, decl_name, f_new_line + value), ""]
 
         # Tabulate non-zero indices.
-        if optimise_parameters["non zero columns"]:
+        if optimise_parameters["eliminate zeros"]:
             if name in name_map:
                 for n in name_map[name]:
                     if inv_name_map[n][1] and inv_name_map[n][1] in new_nzcs:
@@ -551,4 +540,3 @@ def _tabulate_psis(tables, used_psi_tables, inv_name_map, used_nzcs, optimise_pa
                         new_nzcs.remove(inv_name_map[n][1])
     return code
 
-
