@@ -6,7 +6,7 @@ __copyright__ = "Copyright (C) 2009-2010 Kristian B. Oelgaard"
 __license__  = "GNU GPL version 3 or any later version"
 
 # Modified by Anders Logg, 2009.
-# Last changed: 2010-02-15
+# Last changed: 2010-05-18
 
 # UFL modules
 from ufl.classes import Form, Integral, SpatialDerivative
@@ -47,31 +47,48 @@ def compute_integral_ir(domain_type, domain_id, integrals, metadata, form_data,
     ir["prim_idims"] = prim_idims
 
     # Create optimise parameters.
-    # FIXME: Find a way to avoid the redundant options here and in the generator.
+    optimise_parameters = {"eliminate zeros":     False,
+                           "ignore ones":         False,
+                           "remove zero terms":   False,
+                           "optimisation":        False,
+                           "ignore zero tables":  False}
+
     if parameters["optimize"]:
-        # These parameters results in fast code, but compiles slower and there
-        # might still be bugs.
-        optimise_parameters = {"non zero columns": True,
-                            "ignore ones": True,
-                            "remove zero terms": True,
-                            "simplify expressions": True,
-                            "ignore zero tables": True}
-    else:
-        # These parameters should be safe and fast, but result in slow code.
-        optimise_parameters = {"non zero columns": False,
-                            "ignore ones": False,
-                            "remove zero terms": False,
-                            "simplify expressions": False,
-                            "ignore zero tables": False}
+        optimise_parameters["ignore ones"]        = True
+        optimise_parameters["remove zero terms"]  = True
+        optimise_parameters["ignore zero tables"] = True
+
+        # Do not include this in below if/else clause since we want to be
+        # able to switch on this optimisation in addition to the other
+        # optimisations.
+        if "eliminate_zeros" in parameters:
+            optimise_parameters["eliminate zeros"] = True
+
+        if "simplify_expressions" in parameters:
+            optimise_parameters["optimisation"] = "simplify_expressions"
+        elif "precompute_ip_const" in parameters:
+            optimise_parameters["optimisation"] = "precompute_ip_const"
+        elif "precompute_basis_const" in parameters:
+            optimise_parameters["optimisation"] = "precompute_basis_const"
+        # The current default optimisation (for -O) is equal to
+        # '-feliminate_zeros -fsimplify_expressions'.
+        else:
+            # If '-O -feliminate_zeros' was given on the command line, do not
+            # simplify expressions
+            if not "eliminate_zeros" in parameters:
+                optimise_parameters["eliminate zeros"] = True
+                optimise_parameters["optimisation"]    = "simplify_expressions"
 
     # Save the optisation parameters.
-    ir ["optimise_parameters"] = optimise_parameters
+    ir["optimise_parameters"] = optimise_parameters
 
     # Create transformer.
-    if optimise_parameters["simplify expressions"]:
-        transformer = QuadratureTransformerOpt(psi_tables, optimise_parameters)
+    if optimise_parameters["optimisation"]:
+        transformer = QuadratureTransformerOpt(psi_tables, quad_weights, \
+            form_data.geometric_dimension, optimise_parameters)
     else:
-        transformer = QuadratureTransformer(psi_tables, optimise_parameters)
+        transformer = QuadratureTransformer(psi_tables, quad_weights,\
+            form_data.geometric_dimension, optimise_parameters)
 
     # Add tables for weights, name_map and basis values.
     ir["quadrature_weights"]  = quad_weights
@@ -104,6 +121,9 @@ def compute_integral_ir(domain_type, domain_id, integrals, metadata, form_data,
     else:
         error("Unhandled domain type: " + str(domain_type))
 
+    # Save tables map, to extract table names for optimisation option -O.
+    ir["psi_tables_map"] = transformer.psi_tables_map
+
     return ir
 
 def _tabulate_basis(sorted_integrals, domain_type, num_facets):
@@ -257,6 +277,7 @@ def _transform_integrals(transformer, integrals, domain_type):
         if domain_type == "interior_facet":
             integrand = propagate_restrictions(integrand)
         terms = transformer.generate_terms(integrand)
-        transformed_integrals.append((point, terms, transformer.functions, {}))
+        transformed_integrals.append((point, terms, transformer.functions, \
+                                      {}, transformer.coordinate))
     return transformed_integrals
 
