@@ -7,7 +7,7 @@ __license__  = "GNU GPL version 3 or any later version"
 
 # Modified by Peter Brune, 2009
 # Modified by Anders Logg, 2009
-# Last changed: 2010-02-08
+# Last changed: 2010-03-11
 
 # Python modules.
 from numpy import shape
@@ -20,6 +20,7 @@ from ufl.classes import FixedIndex
 from ufl.classes import IntValue
 from ufl.classes import FloatValue
 from ufl.classes import Coefficient
+from ufl.expr import Operator
 
 # UFL Algorithms.
 from ufl.algorithms.printing import tree_format
@@ -36,10 +37,10 @@ from reduce_operations import operation_count
 class QuadratureTransformer(QuadratureTransformerBase):
     "Transform UFL representation to quadrature code."
 
-    def __init__(self, ir, optimise_parameters):
+    def __init__(self, *args):
 
         # Initialise base class.
-        QuadratureTransformerBase.__init__(self, ir, optimise_parameters)
+        QuadratureTransformerBase.__init__(self, *args)
 
     # -------------------------------------------------------------------------
     # Start handling UFL classes.
@@ -241,7 +242,7 @@ class QuadratureTransformer(QuadratureTransformerBase):
             return {(): format["power"](val, expo.value())}
         elif isinstance(expo, FloatValue):
             return {(): format["std power"](val, format["floating point"](expo.value()))}
-        elif isinstance(expo, Coefficient):
+        elif isinstance(expo, (Coefficient, Operator)):
             exp = self.visit(expo)
             return {(): format["std power"](val, exp[()])}
         else:
@@ -269,13 +270,24 @@ class QuadratureTransformer(QuadratureTransformerBase):
         # Get the component
         components = self.component()
 
-        # Safety checks.
+        # Safety check.
         ffc_assert(not operands, "Didn't expect any operands for FacetNormal: " + repr(operands))
-        ffc_assert(len(components) == 1, "FacetNormal expects 1 component index: " + repr(components))
 
-        # We get one component.
-        normal_component = format["normal component"](self.restriction, components[0])
-        self.trans_set.add(normal_component)
+        # Handle 1D as a special case.
+        # FIXME: KBO: This has to change for mD elements in R^n : m < n
+        if self.geo_dim == 1:
+            # Safety check.
+            ffc_assert(len(components) == 0, "FacetNormal in 1D does not expect a component index: " + repr(components))
+            normal_component = format["normal component"](self.restriction, "")
+            self.trans_set.add(normal_component)
+        else:
+
+            # Safety check.
+            ffc_assert(len(components) == 1, "FacetNormal expects 1 component index: " + repr(components))
+
+            # We get one component.
+            normal_component = format["normal component"](self.restriction, components[0])
+            self.trans_set.add(normal_component)
 
         return {():normal_component}
 
@@ -468,10 +480,11 @@ class QuadratureTransformer(QuadratureTransformerBase):
         ffc_assert(len(operands) == 1 and () in operands[0] and len(operands[0]) == 1, \
                    "MathFunctions expect one operand of function type: " + repr(operands))
         # Use format function on value of operand.
+        new_operand = {}
         operand = operands[0]
         for key, val in operand.items():
-            operand[key] = format_function(val)
-        return operand
+            new_operand[key] = format_function(val)
+        return new_operand
 
     # -------------------------------------------------------------------------
     # Helper functions for code_generation()
@@ -495,5 +508,5 @@ class QuadratureTransformer(QuadratureTransformerBase):
         ops = self._count_operations(value)
         used_psi_tables = set([v for k, v in self.psi_tables_map.items()])
 
-        return [value, ops, trans_set, used_points, used_psi_tables]
+        return (value, ops, [trans_set, used_points, used_psi_tables])
 
