@@ -5,12 +5,12 @@ __date__ = "2009-10-13"
 __copyright__ = "Copyright (C) 2009-2010 Kristian B. Oelgaard"
 __license__  = "GNU GPL version 3 or any later version"
 
-# Last changed: 2010-02-08
+# Last changed: 2010-03-11
 
 # Python modules.
 from itertools import izip
 import time
-from numpy import shape
+from numpy import shape, array
 
 # UFL Classes.
 from ufl.classes import MultiIndex, FixedIndex, Index
@@ -21,7 +21,7 @@ from ufl.algorithms import propagate_restrictions, Transformer, tree_format, str
 
 # FFC modules.
 from ffc.log import ffc_assert, error, info
-from ffc.fiatinterface import create_element
+from ffc.fiatinterface import create_element, map_facet_points
 from ffc.mixedelement import MixedElement
 from ffc.cpp import format
 
@@ -37,7 +37,7 @@ class QuadratureTransformerBase(Transformer):
 #class QuadratureTransformerBase(ReuseTransformer):
     "Transform UFL representation to quadrature code."
 
-    def __init__(self, psi_tables, optimise_parameters):
+    def __init__(self, psi_tables, quad_weights, geo_dim, optimise_parameters):
 
         Transformer.__init__(self)
 
@@ -48,17 +48,18 @@ class QuadratureTransformerBase(Transformer):
         self.used_psi_tables = set()
         self.psi_tables_map = {}
         self.used_weights = set()
+        self.quad_weights = quad_weights
         self.used_nzcs = set()
-        self.geo_consts = {}
         self.ip_consts = {}
         self.trans_set = set()
         self.functions = {}
         self.function_count = 0
-        self.geo_dim = 0
+        self.geo_dim = geo_dim
         self.points = 0
         self.facet0 = None
         self.facet1 = None
         self.restriction = None
+        self.coordinate = None
 
         # Stacks.
         self._derivatives = []
@@ -74,6 +75,7 @@ class QuadratureTransformerBase(Transformer):
     def update_facets(self, facet0, facet1):
         self.facet0 = facet0
         self.facet1 = facet1
+        self.coordinate = None
 #        # Reset functions and count everytime we generate a new case of facets.
 #        self.functions = {}
 #        self.function_count = 0
@@ -84,6 +86,7 @@ class QuadratureTransformerBase(Transformer):
 
     def update_points(self, points):
         self.points = points
+        self.coordinate = None
         # Reset functions everytime we move to a new quadrature loop
         self.functions = {}
         self.function_count = 0
@@ -100,7 +103,6 @@ class QuadratureTransformerBase(Transformer):
         print "\nQuadratureTransformer, used_psi_tables:\n", self.used_psi_tables
         print "\nQuadratureTransformer, psi_tables_map:\n", self.psi_tables_map
         print "\nQuadratureTransformer, used_weights:\n", self.used_weights
-        print "\nQuadratureTransformer, geo_consts:\n", self.geo_consts
 
     def component(self):
         "Return current component tuple."
@@ -205,10 +207,6 @@ class QuadratureTransformerBase(Transformer):
         print "\n\nVisiting GeometricQuantity:", repr(o)
         error("This type of GeometricQuantity is not supported (yet).")
 
-    def spatial_coordinate(self, o):
-        print "\n\nVisiting SpatialCoordinate:", repr(o)
-        error("SpatialCoordinate is not supported (yet).")
-
     def lifting_result(self, o):
         print "\n\nVisiting LiftingResult:", repr(o)
         error("LiftingResult (and children) is not supported (yet).")
@@ -277,7 +275,7 @@ class QuadratureTransformerBase(Transformer):
         # Check if basis is already in cache
         basis = self.argument_cache.get((o, components, derivatives, self.restriction), None)
         # FIXME: Why does using a code dict from cache make the expression manipulations blow (MemoryError) up later?
-        if basis is not None and not self.optimise_parameters["simplify expressions"]:
+        if basis is not None and not self.optimise_parameters["optimisation"]:
 #        if basis is not None:
             return basis
 
@@ -364,7 +362,7 @@ class QuadratureTransformerBase(Transformer):
         # Check if function is already in cache
         function_code = self.function_cache.get((o, components, derivatives, self.restriction), None)
         # FIXME: Why does using a code dict from cache make the expression manipulations blow (MemoryError) up later?
-        if function_code is not None and not self.optimise_parameters["simplify expressions"]:
+        if function_code is not None and not self.optimise_parameters["optimisation"]:
 #        if function_code is not None:
             return function_code
 
@@ -444,6 +442,26 @@ class QuadratureTransformerBase(Transformer):
         coefficient = format["coefficient"](o.count(), component)
         return self._create_symbol(coefficient, CONST)
 
+    # -------------------------------------------------------------------------
+    # SpatialCoordinate (geometry.py).
+    # -------------------------------------------------------------------------
+    def spatial_coordinate(self, o, *operands):
+        #print "\n\nVisiting SpatialCoordinate:", repr(o)
+        #print "\n\nVisiting SpatialCoordinate:", repr(operands)
+
+        # Get the component.
+        components = self.component()
+
+        # Safety checks.
+        ffc_assert(not operands, "Didn't expect any operands for FacetNormal: " + repr(operands))
+        ffc_assert(len(components) == 1, " expects 1 component index: " + repr(components))
+
+        # Generate the appropriate coordinate and update tables.
+        coordinate = format["ip coordinates"](self.points, components[0])
+        self._generate_affine_map()
+
+        return self._create_symbol(coordinate, IP)
+
     # -------------------------------------------------------------------------
     # Indexed (indexed.py).
     # -------------------------------------------------------------------------
@@ -523,6 +541,22 @@ class QuadratureTransformerBase(Transformer):
         #print("\n\nVisiting Sin: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))
         return self._math_function(operands, format["sin"])
 
+    def tan(self, o, *operands):
+        #print("\n\nVisiting Tan: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))
+        return self._math_function(operands, format["tan"])
+
+    def acos(self, o, *operands):
+        #print("\n\nVisiting Acos: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))
+        return self._math_function(operands, format["acos"])
+
+    def asin(self, o, *operands):
+        #print("\n\nVisiting Asin: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))
+        return self._math_function(operands, format["asin"])
+
+    def atan(self, o, *operands):
+        #print("\n\nVisiting Atan: " + repr(o) + "with operands: " + "\n".join(map(repr,operands)))
+        return self._math_function(operands, format["atan"])
+
     # -------------------------------------------------------------------------
     # PositiveRestricted and NegativeRestricted (restriction.py).
     # -------------------------------------------------------------------------
@@ -628,20 +662,82 @@ class QuadratureTransformerBase(Transformer):
 
         f_nzc = format["nonzero columns"](0).split("0")[0]
 
-        # Loop code and add weight and scale factor.
+        # Loop code and add weight and scale factor to value and sort after
+        # loop ranges.
         new_terms = {}
         for key, val in terms.items():
             # If value was zero continue.
             if val is None:
                 continue
-
-            # Create data and insert.
-            data = self._create_entry_data(val)
+            # Create data.
+            value, ops, sets = self._create_entry_data(val)
+            # Extract nzc columns if any and add to sets.
             used_nzcs = set([int(k[1].split(f_nzc)[1].split("[")[0]) for k in key if f_nzc in k[1]])
-            data.append(used_nzcs)
-            new_terms[key] = data
+            sets.append(used_nzcs)
+
+            # Create loop information and entry from key info and insert into dict.
+            loop, entry = self._create_loop_entry(key)
+            if not loop in new_terms:
+                sets.append({})
+                new_terms[loop] = [sets, [(entry, value, ops)]]
+            else:
+                for i, s in enumerate(sets):
+                    new_terms[loop][0][i].update(s)
+                new_terms[loop][1].append((entry, value, ops))
         return new_terms
 
+    def _create_loop_entry(self, key):
+
+        # TODO: Verify that test and trial functions will ALWAYS be rearranged to 0 and 1.
+        indices = {-2: format["first free index"], -1: format["second free index"],
+                    0: format["first free index"],  1: format["second free index"]}
+
+        # Create appropriate entries.
+        # FIXME: We only support rank 0, 1 and 2.
+        entry = ""
+        loop = ()
+        if len(key) == 0:
+            entry = "0"
+        elif len(key) == 1:
+            key = key[0]
+            # Checking if the basis was a test function.
+            # TODO: Make sure test function indices are always rearranged to 0.
+            ffc_assert(key[0] == -2 or key[0] == 0, \
+                        "Linear forms must be defined using test functions only: " + repr(key))
+            index_j, entry, range_j, space_dim_j = key
+            loop = ((indices[index_j], 0, range_j),)
+            if range_j == 1 and self.optimise_parameters["ignore ones"]:
+                loop = ()
+        elif len(key) == 2:
+            # Extract test and trial loops in correct order and check if for is legal.
+            key0, key1 = (0, 0)
+            for k in key:
+                ffc_assert(k[0] in indices, \
+                "Bilinear forms must be defined using test and trial functions (index -2, -1, 0, 1): " + repr(k))
+                if k[0] == -2 or k[0] == 0:
+                    key0 = k
+                else:
+                    key1 = k
+            index_j, entry_j, range_j, space_dim_j = key0
+            index_k, entry_k, range_k, space_dim_k = key1
+
+            loop = []
+            if not (range_j == 1 and self.optimise_parameters["ignore ones"]):
+                loop.append((indices[index_j], 0, range_j))
+            if not (range_k == 1 and self.optimise_parameters["ignore ones"]):
+                loop.append((indices[index_k], 0, range_k))
+            entry = format["add"]([format["mul"]([entry_j, str(space_dim_k)]), entry_k])
+            loop = tuple(loop)
+        else:
+            error("Only rank 0, 1 and 2 tensors are currently supported: " + repr(key))
+        # Generate the code line for the entry.
+        # Try to evaluate entry ("3*6 + 2" --> "20").
+        try:
+            entry = str(eval(entry))
+        except:
+            pass
+        return loop, entry
+
     # -------------------------------------------------------------------------
     # Helper functions for transformation of UFL objects in base class
     # -------------------------------------------------------------------------
@@ -696,18 +792,8 @@ class QuadratureTransformerBase(Transformer):
         if component:
             local_offset = component - local_comp
 
-        # Set geo_dim.
-        # TODO: All terms REALLY have to be defined on cell with the same
-        # geometrical dimension so only do this once and exclude the check?
-        geo_dim = ufl_function.element().cell().geometric_dimension()
-        if self.geo_dim:
-            ffc_assert(geo_dim == self.geo_dim, \
-                       "All terms must be defined on cells with the same geometrical dimension.")
-        else:
-            self.geo_dim = geo_dim
-
         # Generate FFC multi index for derivatives.
-        multiindices = FFCMultiIndex([range(geo_dim)]*len(derivatives)).indices
+        multiindices = FFCMultiIndex([range(self.geo_dim)]*len(derivatives)).indices
 
         return (component, local_comp, local_offset, ffc_element, quad_element, transformation, multiindices)
 
@@ -909,6 +995,41 @@ class QuadratureTransformerBase(Transformer):
                 ffc_assert(data[1] == loop_index_range, "Index ranges does not match." + repr(data[1]) + repr(loop_index_range))
         return function_name
 
+    def _generate_affine_map(self):
+        """Generate psi table for affine map, used by spatial coordinate to map
+        integration point to physical element."""
+
+        # TODO: KBO: Perhaps it is better to create a fiat element and tabulate
+        # the values at the integration points?
+        f_FEA = format["affine map table"]
+        f_ip  = format["integration points"]
+
+        affine_map = {1: lambda x: [1.0 - x[0],               x[0]],
+                      2: lambda x: [1.0 - x[0] - x[1],        x[0], x[1]],
+                      3: lambda x: [1.0 - x[0] - x[1] - x[2], x[0], x[1], x[2]]}
+
+        num_ip = self.points
+        w, points = self.quad_weights[num_ip]
+        if not self.facet0 is None:
+            points = map_facet_points(points, self.facet0)
+            name = f_FEA(num_ip, self.facet0)
+        else:
+            name = f_FEA(num_ip, 0)
+
+        if name not in self.unique_tables:
+            vals = []
+            for p in points:
+                vals.append(affine_map[len(p)](p))
+            self.unique_tables[name] = array(vals)
+        if self.coordinate is None:
+            ip = 1
+            r = None
+            if num_ip > 1:
+                ip = f_ip
+            if self.facet1 is not None:
+                r = "+"
+            self.coordinate = [name, self.geo_dim, ip, r]
+
     # -------------------------------------------------------------------------
     # Helper functions for code_generation()
     # -------------------------------------------------------------------------
