@@ -5,7 +5,7 @@ __date__ = "2009-07-12"
 __copyright__ = "Copyright (C) 2009-2010 Kristian B. Oelgaard"
 __license__  = "GNU GPL version 3 or any later version"
 
-# Last changed: 2010-02-09
+# Last changed: 2010-03-11
 
 # FFC modules.
 from ffc.log import error
@@ -19,7 +19,7 @@ from symbolics import create_fraction
 from expr import Expr
 
 class Symbol(Expr):
-    __slots__ = ("v", "base_expr", "base_op")
+    __slots__ = ("v", "base_expr", "base_op", "exp")
     def __init__(self, variable, symbol_type, base_expr=None, base_op=0):
         """Initialise a Symbols object, it derives from Expr and contains
         the additional variables:
@@ -43,6 +43,7 @@ class Symbol(Expr):
         # ops = base_expr.ops() + base_ops = 2 + 1 = 3
         self.base_expr = base_expr
         self.base_op = base_op
+        self.exp = None
 
         # If type of the base_expr is lower than the given symbol_type change type.
         # TODO: Should we raise an error here? Or simply require that one
@@ -64,7 +65,11 @@ class Symbol(Expr):
     # Print functions.
     def __str__(self):
         "Simple string representation which will appear in the generated code."
-        return self.v
+        if self.base_expr is None:
+            return self.v
+        elif self.exp is None:
+            return self.v(str(self.base_expr))
+        return self.v(str(self.base_expr), self.exp)
 
     # Binary operators.
     def __add__(self, other):
