@@ -1,9 +1,20 @@
+"""This script compiles and verifies the output for all form files
+found in the 'demo' directory. The verification is performed in two
+steps. First, the generated code is compared with stored references.
+Then, the output from all functions in the generated code is compared
+with stored reference values.
+
+This script can also be used for benchmarking tabulate_tensor for all
+form files found in the 'bench' directory. To run benchmarks, use the
+option -b.
+"""
+
 __author__ = "Anders Logg, Kristian B. Oelgaard and Marie E. Rognes"
 __date__ = "2010-01-21"
 __copyright__ = "Copyright (C) 2010 " + __author__
 __license__  = "GNU GPL version 3 or any later version"
 
-# Last changed: 2010-02-12
+# Last changed: 2010-05-12
 
 # FIXME: Need to add many more test cases. Quite a few DOLFIN
 # FIXME: forms failed after the FFC tests passed. Also need to
@@ -19,6 +30,7 @@ from ufctest import generate_test_code
 tolerance = 1e-9
 output_directory = "output"
 demo_directory = "../../../demo"
+bench_directory = "../../../bench"
 
 # Global log file
 logfile = None
@@ -47,30 +59,35 @@ def clean_output():
         shutil.rmtree(output_directory)
     os.mkdir(output_directory)
 
-def generate_test_cases():
+def generate_test_cases(bench):
     "Generate form files for all test cases."
 
     begin("Generating test cases")
 
-    # Copy demos files
-    demo_files = [f for f in os.listdir(demo_directory) if f.endswith(".ufl")]
-    demo_files.sort()
-    for f in demo_files:
-        shutil.copy("%s/%s" % (demo_directory, f), ".")
-    info_green("Copied %d demo files" % len(demo_files))
+    # Copy form files
+    if bench:
+        form_directory = bench_directory
+    else:
+        form_directory = demo_directory
+    form_files = [f for f in os.listdir(form_directory) if f.endswith(".ufl")]
+    form_files.sort()
+    for f in form_files:
+        shutil.copy("%s/%s" % (form_directory, f), ".")
+    info_green("Found %d form files" % len(form_files))
 
     # Generate form files for forms
-    info("Generating form files for extra demo forms: Not implemented")
+    info("Generating form files for extra forms: Not implemented")
 
     # Generate form files for elements
-    from elements import elements
-    info("Generating form files for extra elements (%d elements)" % len(elements))
-    for (i, element) in enumerate(elements):
-        open("X_Element%d.ufl" % i, "w").write("element = %s" % element)
+    if not bench:
+        from elements import elements
+        info("Generating form files for extra elements (%d elements)" % len(elements))
+        for (i, element) in enumerate(elements):
+            open("X_Element%d.ufl" % i, "w").write("element = %s" % element)
 
     end()
 
-def generate_code():
+def generate_code(args):
     "Generate code for all test cases."
 
     # Get a list of all files
@@ -83,7 +100,7 @@ def generate_code():
     for f in form_files:
 
         # Generate code
-        ok = run_command("ffc -d -f precision=8 -fconvert_exceptions_to_warnings %s" % f)
+        ok = run_command("ffc %s -d -f precision=8 -fconvert_exceptions_to_warnings %s" % (" ".join(args), f))
 
         # Check status
         if ok:
@@ -128,7 +145,7 @@ def validate_code():
 
     end()
 
-def build_programs():
+def build_programs(bench):
     "Build test programs for all test cases."
 
     # Get a list of all files
@@ -137,15 +154,25 @@ def build_programs():
 
     begin("Building test programs (%d header files found)" % len(header_files))
 
+    # Set compiler options
+    if bench > 0:
+        info("Benchmarking activated")
+        # Takes too long to build with -O2
+        #compiler_options = "`pkg-config --cflags ufc-1` -Wall -Werror -O2"
+        compiler_options = "`pkg-config --cflags ufc-1` -Wall -Werror"
+    else:
+        compiler_options = "`pkg-config --cflags ufc-1` -Wall -Werror -g"
+    info("Compiler options: %s" % compiler_options)
+
     # Iterate over all files
     for f in header_files:
 
         # Generate test code
-        filename = generate_test_code(f)
+        filename = generate_test_code(f, bench)
 
         # Compile test code
         prefix = f.split(".h")[0]
-        command = "g++ `pkg-config --cflags ufc-1` -Wall -Werror -g -o %s.bin %s.cpp" % (prefix, prefix)
+        command = "g++ %s -o %s.bin %s.cpp" % (compiler_options, prefix, prefix)
         ok = run_command(command)
 
         # Check status
@@ -257,18 +284,29 @@ def main(args):
     # Enter output directory
     os.chdir(output_directory)
 
-    # Generate test cases
-    generate_test_cases()
+    # Check command-line arguments
+    bench = "--bench" in args
+    fast = "--fast" in args
+    args = [arg for arg in args if not arg in ("--bench", "--fast")]
 
-    # Generate and validate code
-    generate_code()
-    validate_code()
+    # Generate test cases
+    generate_test_cases(bench)
+
+    # Generate code
+    generate_code(args)
+
+    # Validate code
+    if not bench:
+        validate_code()
 
     # Build, run and validate programs
-    if "--fast" in args or "-f" in args:
+    if fast:
         info("Skipping program validation")
+    elif bench:
+        build_programs(bench)
+        run_programs()
     else:
-        build_programs()
+        build_programs(bench)
         run_programs()
         validate_programs()
 
